from datetime import datetime, timedelta
from typing import List, Optional

import feedparser
from pick import pick

from .dates import get_pretty_date_str, parse_date


class Article(feedparser.FeedParserDict):
    """Represents a single article in a feed"""

    def get_parsed_date(self: feedparser.FeedParserDict) -> datetime:
        """Return a datetime object parsed from the string date value"""
        raw_date = self.get_raw_date()
        return parse_date(raw_date)

    def get_raw_date(self: feedparser.FeedParserDict) -> str:
        """Return the raw string date as set up in the entry"""
        default = str(datetime.utcnow() - timedelta(days=30))
        return self.get("published") or self.get("updated") or default

    def get_str_date(self: feedparser.FeedParserDict) -> str:
        """Return a nicely formatted string from the date in the entry"""
        raw_date = self.get_raw_date()
        return get_pretty_date_str(raw_date)


class Feed(feedparser.FeedParserDict):
    """Represents a single, feedparser-parsed RSS feed"""

    def __init__(self, feed_url: str):
        """Constructor"""
        super().__init__(feedparser.parse(feed_url))

    def find_unread_articles_from_user(self) -> List[Article]:
        """Ask the user what was the last article they have already read, and return all the newer ones"""
        last_read_index = self.find_last_read_article()
        return [Article(entry) for entry in self.entries[:last_read_index]]

    def find_last_read_article(self) -> int:
        """Find the last article from the feed that the user has already read, and return its index"""
        articles = [Article(entry) for entry in self.entries]
        options = [f"{article.title} ({article.get_str_date()})" for article in articles]
        options.append("I haven't read any of these :(")
        title = "Please choose the *last* article from this feed that you've already read"
        _, index = pick(options, title)
        return index

    def find_unread_articles_from_date(self, last_updated: datetime) -> List[Article]:
        """Find all the new articles since `last_updated`"""
        # Some fancy walrus operator fun, because why not?
        return [article for entry in self.entries if (article := Article(entry)).get_parsed_date() > last_updated]

    def get_unread_articles(self, last_updated: Optional[datetime]) -> List[Article]:
        """
        Return all the new articles

            * either since the last time r2k has ran, or
            * if this is the first time r2k is running for this feed, by asking the user which was the last article
            from the feed that they have already read, and only sending them the ones that came after
        """
        if last_updated:
            unread_articles = self.find_unread_articles_from_date(last_updated)
        else:
            unread_articles = self.find_unread_articles_from_user()

        # Reverse, as we want the oldest articles first
        unread_articles.reverse()
        return unread_articles



from __future__ import annotations

import sys
from time import sleep
from types import TracebackType
from typing import Optional, Type

import docker
import requests
from docker.errors import APIError as DockerAPIError
from docker.models.containers import Container
from requests.exceptions import ConnectionError

from r2k.cli import logger

CONTAINER_NAME = "mercury-parser-api"
MERCURY_PORT = 3000
BASE_MERCURY_URL = f"http://localhost:{MERCURY_PORT}/parser"
CONNECTION_ATTEMPTS = 10


class MercuryParser:
    """
    Represents the gateway to the Mercury Parser API

    Relies on https://hub.docker.com/r/wangqiru/mercury-parser-api
    """

    def __init__(self) -> None:
        """Constructor"""
        self.container: Optional[Container] = None
        self.client: docker.DockerClient = docker.from_env()

    def run_mercury_container(self) -> Container:
        """Launch a new mercury-parser docker container"""
        logger.debug("Launching a new mercury-parser Docker container...")
        self.container = self.client.containers.run(
            "wangqiru/mercury-parser-api:latest",
            detach=True,
            ports={f"{MERCURY_PORT}/tcp": MERCURY_PORT},
            name=CONTAINER_NAME,
        )
        return self.container

    def clean_existing_containers(self) -> None:
        """Remove any existing mercury parser API containers"""
        all_containers = self.client.containers.list(all=True, sparse=True)
        for container in all_containers:
            if container.attrs["Names"] == [f"/{CONTAINER_NAME}"]:
                logger.debug("Found an existing container with the same name...")
                self.remove_container(container)

    def remove_container(self, container: Optional[Container] = None) -> None:
        """Stop and remove a docker container"""
        container = container or self.container
        if container:
            logger.debug("Stopping container...")
            container.stop()
            logger.debug("Removing container...")
            container.remove()

    def __enter__(self) -> MercuryParser:
        """
        Context manager __enter__ for MercuryParser

            1. Clean any existing containers
            2. Spin a local, dockerized version of the mercury parser API
            3. Validate it's up
            4. Return the MercuryParser instance
        """
        self.clean_existing_containers()
        self.run_mercury_container()
        self.validate_container_is_up()
        return self

    def __exit__(
        self, exc_type: Optional[Type[BaseException]], exc_val: Optional[BaseException], exc_tb: Optional[TracebackType]
    ) -> bool:
        """
        Context manager __exit__ for MercuryParser

            1. Remove the container
            2. Stop the program for any expected errors
        """
        self.remove_container()

        if exc_val:
            if isinstance(exc_val, (DockerAPIError, ConnectionError)):
                logger.error("Could not connect to Docker. Run with -v to get more details")
                logger.debug(f"Error info:\n{exc_val}")
                sys.exit(1)
            else:
                raise exc_val
        return True

    @staticmethod
    def validate_container_is_up() -> None:
        """Try to connect to the mercury parser service several times. Quit app if not successful"""
        errors = set()
        logger.debug(f"Launched container at {BASE_MERCURY_URL}. Validating it's up...")
        while retries := CONNECTION_ATTEMPTS:
            try:
                requests.get(BASE_MERCURY_URL)
                logger.debug("Connected!")
                return
            except ConnectionError as e:
                errors.add(e)
            sleep(1)
            retries -= 1

        logger.error("Could not connect to the mercury-parser Docker container")
        if errors:
            errors_str = "\n".join(str(error) for error in errors)
            logger.error(f"Error info:{errors_str}")
        sys.exit(1)

    def parse(self, url: str) -> dict:
        """
        Parse a single URL with the Mercury Parser and return the result
        """
        result = self.get_parsed_doc(url)
        if result.get("error"):
            error_msg = result.get("message", "Unknown")
            logger.error(f"Failed to parse {url}\nError: {error_msg}")
            return {}
        return result

    @staticmethod
    def get_parsed_doc(url: str) -> dict:
        """Make an HTTP call to the mercury API and get the parsed document"""
        full_url = f"{BASE_MERCURY_URL}?url={url}"
        logger.debug("Parsing article with Mercury Parser...")
        logger.debug(f"Sending request to {full_url}")
        result = requests.get(full_url).json()
        logger.debug("Finished parsing")
        return result



from typing import Any, Optional

import click

from .cli_utils import is_verbose, no_ansi


def style(msg: str, fg: str, bold: bool) -> str:
    """
    Styles a text with ANSI styles and returns the new string

    If the --no-ansi flag was passed, no styles are applied
    """
    if no_ansi():
        return msg
    else:
        return click.style(msg, fg=fg, bold=bold)


def secho(msg: Any, fg: str, bold: bool = False) -> None:
    """Similar to click.secho"""
    if not isinstance(msg, (str, bytes)):
        msg = str(msg)
    click.echo(style(msg, fg=fg, bold=bold))


def warning(msg: Any) -> None:
    """click.echo in a yellow color"""
    secho(f"WARNING: {msg}", "yellow")


def error(msg: Any) -> None:
    """click.echo in a red color"""
    secho(f"ERROR: {msg}", "red")


def info(msg: Any) -> None:
    """click.echo in a blue color"""
    secho(msg, "blue")


def notice(msg: Any) -> None:
    """click.echo in a green color"""
    secho(msg, "green")


def log(msg: Any) -> None:
    """click.echo in a white color"""
    secho(msg, "white")


def debug(msg: Any) -> None:
    """click.echo in a white color, only if the --verbose flag was set"""
    if is_verbose():
        log(msg)


def prompt(
    text: str, hide_input: bool = False, fg: str = "cyan", default: Any = None, type: Optional[click.ParamType] = None
) -> Any:
    """Ask the user to enter a value, and return the result"""
    text = style(text, fg=fg, bold=True)
    return click.prompt(text, hide_input=hide_input, default=default, type=type)


def confirm(text: str, default: bool = True, fg: str = "cyan") -> bool:
    """Ask the user for confirmation and return the bool result"""
    text = style(text, fg=fg, bold=True)
    return click.confirm(text, default=default)



import click

from .feed_add import feed_add
from .feed_import import feed_import
from .feed_remove import feed_remove
from .feed_show import feed_show


@click.group()
def feed() -> None:
    """Interact with the RSS feeds."""
    pass


feed.add_command(feed_import)
feed.add_command(feed_show)
feed.add_command(feed_remove)
feed.add_command(feed_add)



import click
import yaml

from r2k.cli import cli_utils, logger
from r2k.config import config


@click.command("show")
@cli_utils.config_path_option()
def feed_show() -> None:
    """List all existing RSS feeds."""
    if config.feeds:
        feeds = yaml.safe_dump(config.feeds)  # Just here to make the output nicer
        logger.info("Here are the existing feeds:")
        logger.secho(feeds, fg="white", bold=False)
    else:
        logger.info("There are no feeds available.\n" "Add more by running `r2k feeds import/add`")



from datetime import datetime
from typing import Union

from dateutil.parser import parse as parse_date


def get_pretty_date_str(date: Union[str, datetime], show_time: bool = False, show_year: bool = True) -> str:
    """Return a nice representation of a date"""
    if isinstance(date, str):
        date = parse_date(date)

    # Only day and Month
    template = "%d %B"

    if show_year:
        template = f"{template}, %Y"

    # Add the time
    if show_time:
        template = f"%H:%M {template}"
    return date.strftime(template)


def now() -> datetime:
    return datetime.utcnow()
